from sklearn.linear_model          import LogisticRegression
from sklearn.neural_network        import MLPClassifier
from sklearn.neighbors             import KNeighborsClassifier
from sklearn.svm                   import SVC
from sklearn.gaussian_process      import GaussianProcessClassifier
from sklearn.tree                  import DecisionTreeClassifier
from sklearn.ensemble              import RandomForestClassifier, AdaBoostClassifier
from sklearn.naive_bayes           import GaussianNB
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis

valid_models = {name : obj for name, obj in globals().items() if 'sklearn' in repr(obj)}

def get_model(name):
    if name not in valid_models:
        raise Exception()
    return valid_models.get(name)    



def train(output_dir, meta, model_config, feature_selection, num_features, min_count, target_subset, metric, aggregate_func):

    import os
    import json

    command = 'genolearn train'

    from genolearn.utils                 import create_log
    from genolearn.models.classification import get_model
    from genolearn.models                import grid_predictions
    from genolearn.dataloader            import DataLoader
    from genolearn.logger                import msg, Writing

    import warnings
    import numpy as np
    import os
    import json
    import pickle


    os.makedirs(output_dir)

    warnings.simplefilter("ignore")
    os.environ["PYTHONWARNINGS"] = "ignore"


    # data_config, model_config = map(check_config, (data_config, model_config))

    # with open(model_config) as f:
    #     model_config = json.load(model_config)


    # print_dict('executing "train.py" with parameters', params)

    with open(os.path.join('model', model_config)) as file:
        model_config = json.load(file)
        model        = model_config.pop('model')

    kwargs     = {key : val for key, val in model_config.items() if isinstance(val, list)}
    common     = {key : val for key, val in model_config.items() if key not in kwargs}

    dataloader = DataLoader(meta)
    selection  = dataloader.load_feature_selection(feature_selection).argsort()

    Model   = get_model(model)
    
    outputs, params = grid_predictions(dataloader, Model, selection, num_features, common, min_count, target_subset, metric, aggregate_func, **kwargs)
        
    model, predict, *probs = outputs.pop('best')

    target  = outputs['target']

    npz     = os.path.join(output_dir, 'results.npz')
    pkl     = os.path.join(output_dir, 'model.pickle')
    csv     = os.path.join(output_dir, 'predictions.csv')
    js      = os.path.join(output_dir, 'params.json')
    enc     = os.path.join(output_dir, 'encoding.json')

    dump    = np.c_[outputs['identifiers'], np.array([target, predict]).T]
    headers = ['identifier', 'target', 'predict']

    if probs:
        for i, label in enumerate(dataloader._encoder):
            dump = np.c_[dump, probs[0][:,i].astype(str)]
            headers.append(f'P({label})')

    dump = ','.join(headers) + '\n' + '\n'.join(','.join(row) for row in dump)

    with open(csv, 'w') as f:
        f.write(dump)

    with Writing(npz, inline = True):
        np.savez_compressed(npz, **outputs)

    with Writing(pkl, inline = True):
        with open(pkl, 'wb') as f:
            pickle.dump(model, f)

    with Writing(js, inline = True):
        with open(js, 'w') as f:
            f.write(json.dumps(params, indent = 4))

    with open(enc, 'w') as f:
        f.write(json.dumps(dataloader._encoder, indent = 4))

    create_log('train', output_dir)

    msg(f'executed "{command}"')



def _analyse(meta, min_count, proportion):
    
    import pandas as pd

    check   = set(meta['group']) != {'Train', 'Test'}
    columns = list(meta['group']) + ['Train' ,'Test'] if check else ['Train' ,'Test']
    count   = pd.DataFrame(index = meta['targets'], columns = columns, data = 0)

    if check:
        for group in meta['group']:
            for identifier in meta['group'][group]:
                target = meta['search'][identifier]
                count.loc[target, group] += 1

    for key in ['Train', 'Test']:
        for group in meta[key]:
            for identifier in meta['group'][group]:
                count.loc[meta['search'][identifier], key] += 1

    count['Global']    = count['Train'] + count['Test']
    count.loc['Total'] = count.sum(axis = 0).copy()

    suggested_subset = count.index[:-1][count['Train'].values[:-1] >= min_count]

    if len(suggested_subset) == len(count.index):
        suggested_subset = None

    if proportion:
        count.iloc[:-1] /= count.iloc[:-1].sum(axis = 0)
        count.iloc[-1]   = count.iloc[-1] / count.iloc[-1,-1]
        count *= 100
        return count.round(2).applymap(lambda value : f'{value:6.2f}'), suggested_subset
    else:
        return count.applymap(int).applymap(lambda value : f'{value:6,d}'), suggested_subset
    
def analyse(meta, min_count, proportion):
    import json

    with open(meta) as f:
        meta = json.load(f)

    count, subset = _analyse(meta, min_count, proportion)

    maxlen1 = max(map(len, count.index))
    maxlen2 = max([max(map(len, count.values.flatten())), max(map(len, count.columns))])

    print(' ' * maxlen1 + ' | ' + ' | '.join(f'{col:>{maxlen2}s}' for col in count.columns))
    print('-' * (maxlen1 + 1) + '+' + '+'.join('-' * (maxlen2 + 2) for _ in range(count.shape[1])))
    for target in count.index:
        print(f'{target:{maxlen1}s} | ' + ' | '.join(list(count.loc[target])))

    if subset is not None:
        print('\nsuggested target subset:', ', '.join(subset))

def _load_df(meta):
    import pandas as pd
    import json

    with open(meta) as f:
        meta = json.load(f)

    g  = []
    t  = []
    for identifier in meta['identifiers']:
        t.append(meta['search'][identifier])
        for group, identifiers in meta['group'].items():
            if identifier in identifiers:
                g.append(group)
                continue

    df = pd.DataFrame()
    df['identifier'] = meta['identifiers']
    df[' group']     = g
    df[' target']    = t
    df.index        += 1
    return df

def head(meta, num):
    df = _load_df(meta)
    print(df.head(num))
    

def tail(meta, num):
    df = _load_df(meta)
    print(df.tail(num))

def sample(meta, num):
    df = _load_df(meta)
    print(df.sample(num))



def evaluate(model, data_config, feature_selection, values, encoder, num_features, output):

    from   genolearn.logger import msg

    from   genolearn.dataloader import DataLoader

    import joblib
    import pandas as pd

    model       = joblib.load(model)
    dataloader  = DataLoader(**data_config)

    features = dataloader.load_feature_selection(feature_selection).argsort()[:num_features]

    X   = dataloader.load_X(*values, features = features)

    dataloader._encoder = encoder

    df  = pd.DataFrame(index = dataloader._check_identifiers(values))
    df.index.name = 'identifier'

    df['hat'] = dataloader.decode(model.predict(X))
    
    if hasattr(model, 'predict_proba'):
        prob = model.predict_proba(X)
        for i, name in enumerate(encoder):
            df[f'P({name})'] = prob[:,i]
    
    df.to_csv(output)

    msg('executed "genolearn evaluate"')



from   genolearn import working_directory
import os
import json
import scipy.sparse
import numpy  as np
import gzip

class DataLoader():
    """
    DataLoader Class

    Parameters
    ----------
        meta_file   : str
            Preprocessed metadata within ``preprocess_dir``/meta
        working_dir : str
            Path to working directory

    """
    def __init__(self, meta_file, working_dir = working_directory):
        
        if working_dir is None:
            raise Exception('execute GenoLearn\'s setup command first')

        if not os.path.exists(working_dir) or '.genolearn' not in os.listdir(working_dir):
            raise Exception(f'"{working_dir}" not a valid working directory')

        path  = os.path.join(working_dir, 'meta')
        if not os.path.exists(path):
            raise Exception('no preprocessed metadata to load')
        
        valid = os.listdir(path)

        if len(valid) == 0:
            raise Exception(f'no preprocessed metadata in {path}')
            
        if meta_file not in valid:
            raise Exception(f'"{meta_file}" not found in "{path}" - expect one of' + '\n  •'.join([''] + valid))

        self.dense          = os.path.exists(os.path.join(working_dir, 'preprocess', 'dense'))
        self.working_dir    = working_dir
        self.preprocess_dir = os.path.join(working_dir, 'preprocess')
        self.data_dir       = os.path.join(self.preprocess_dir, 'dense' if self.dense else 'sparse')
        
        with open(os.path.join(working_dir, 'meta', meta_file)) as f:
            self.meta = json.load(f)
            self.c    = len(set(self.meta['targets']))

        with open(os.path.join(self.preprocess_dir, 'info.json')) as f:
            info = json.load(f)
            self.n, self.m = info['n'], info['m']

    def _load_X(self, identifier, features):

        npz = os.path.join(self.data_dir, f'{identifier}.npz')

        if self.dense:
            arr, = np.load(npz).values()
        else:
            arr  = scipy.sparse.load_npz(npz)

        if features is not None:
            if self.dense:
                arr = arr[features]
            else:
                arr = arr[:,features]
        
        return arr

    def _load_Y(self, identifier):
        return self.meta['search'][identifier]

    def _check_identifiers(self, identifiers):
        ret = []
        
        for identifier in identifiers:

            # unlabelled data
            if identifier == 'unlabelled':
                for identifier in os.listdir(self.data_dir):
                    identifier = identifier.replace('.npz', '')
                    if identifier not in self.meta['identifiers']:
                        ret.append(identifier)
                    
            # check if it is a group and append entire associated identifiers
            elif identifier in self.meta['group']:
                ret += self.meta['group'][identifier]
            
            # check if it is Train / Test and append entire associated identifiers
            elif identifier in ['Train', 'Test']:
                for group in self.meta[identifier]:
                    ret += self.meta['group'][group]

            # append the identifier as is
            else:
                ret.append(identifier)

        return ret

    def load_X(self, *identifiers, features = None, dtype = np.uint16):
        r"""
        loads all observations with associated ``identifiers``. If ``features`` is provided, loads only
        those feature values. If ``dtype`` is provided, tries to convert the ``dtype`` provided.
        Returns
        -------
            X : numpy.ndarray or scipy.sparse.csr_matrix
                Defining :math:`n = |\text{identifiers}|`  and :math:`m` as the number of genome
                sequences identified during the preprocessing stage or :math:`|\text{features}|` if ``features``
                was provided, then, :math:`X\in\mathbb{Z}^{n,m}`. 
        """
        identifiers = self._check_identifiers(identifiers)
        m           = self.m if features is None else len(features)
        if self.dense:
            X = np.zeros((len(identifiers), m), dtype = dtype)
            for i, identifier in enumerate(identifiers):
                X[i] = self._load_X(identifier, features)
        else:
            X = scipy.sparse.vstack([self._load(identifier, features) for identifier in identifiers], dtype = dtype)
        return X

    def load_Y(self, *identifiers):
        """
        Returns
        -------
        Y : numpy.ndarray
        """
        identifiers = self._check_identifiers(identifiers)
        return np.array(list(map(self._load_Y, identifiers)))

    def load_train_test_identifiers(self, min_count = 0, target_subset = None):
        """
        loads train and test identifiers.
        Returns
        -------
            train_identifiers : numpy.ndarray
            test_identifiers  : numpy.ndarray
        """
        train   = self._check_identifiers(self.meta['Train'])
        test    = self._check_identifiers(self.meta['Test'])

        y_train = self.load_Y(*train)
        y_test  = self.load_Y(*test)

        unique, arg = np.unique(y_train, return_inverse = True)
        dummies     = np.eye(len(unique))[arg]

        label_counts = dummies.sum(axis = 0)
        labels       = unique[label_counts >= min_count]

        if target_subset:
            labels = [label for label in labels if label in target_subset]
        
        self._encoder = {label : i for i, label in enumerate(labels)}

        identifiers_train = np.array(train)[np.isin(y_train, labels)]
        identifiers_test  = np.array(test )[np.isin(y_test , labels)]

        return identifiers_train, identifiers_test

    def load_train_test(self, features = None, min_count = 0, target_subset = None, dtype = np.uint16):
        """
        using the method ``load_train_test_identifiers`` returns train and test data for supervised learning.
        Returns
        -------
            X_train : load_X(train_identifiers, features = features, sparse = sparse)
            Y_train : load_Y(train_identifiers)
            X_test  : load_X(test_identifiers, features = features, sparse = sparse)
            Y_test  : load_Y(test_identifiers)
        """
        self.identifiers_train, self.identifiers_test = self.load_train_test_identifiers(min_count, target_subset)

        Y_train = self.encode(self.load_Y(*self.identifiers_train))
        Y_test  = self.encode(self.load_Y(*self.identifiers_test ))

        X_train = self.load_X(*self.identifiers_train, features = features, dtype = dtype)
        X_test  = self.load_X(*self.identifiers_test , features = features, dtype = dtype)

        return X_train, Y_train, X_test, Y_test
    
    def encode(self, Y):
        return np.vectorize(lambda value : self._encoder[value])(Y)

    def decode(self, Y):
        decoder = {value : key for key, value in self._encoder.items()}
        return np.vectorize(lambda value : decoder[value])(Y)

    def features(self, indices = None):
        if '_features' not in self.__dict__:
            with gzip.open(os.path.join(self.preprocess_dir, 'features.txt.gz')) as g:
                self._features = g.read().decode().split()
        return self._features if indices is None else [self._features[i] for i in indices]

    def load_feature_selection(self, file):
        ret, = np.load(os.path.join(self.working_dir, 'feature-selection', file), allow_pickle = True).values()
        return ret

    def generator(self, *identifiers, features = None, force_dense = False, force_sparse = False):
        """
        Iteratively yields an x, y pair from the method ``load``.
        """
        identifiers = self._check_identifiers(identifiers)

        for identifier in identifiers:
            X, Y = self._load_X(identifier, features), self.load_Y(identifier)[0]
            if force_sparse:
                if isinstance(X, np.ndarray):
                    X = scipy.sparse.csr_matrix(X.reshape(1, -1))
            if force_dense:
                if not isinstance(X, np.ndarray):
                    X = X.A.flatten()
            yield X, Y


"""
genolearn doc

"""

import os

__version__ = '0.0.8'

ls = os.listdir()
wd = True if '.genolearn' in ls else False

path = os.path.join(os.path.dirname(__file__), 'wd')
if wd:
    working_directory = os.path.abspath('.')
    with open(path, 'w') as f:
        f.write(working_directory)
elif os.path.exists(path):
    with open(path) as f:
        working_directory = f.read()
else:
    working_directory = None

ls = os.listdir(working_directory if working_directory and os.path.exists(working_directory) else '.')

def listdir(path = '.'):
    if working_directory and os.path.exists(working_directory):
        path = os.path.join(working_directory, path)
        return os.listdir(path) if os.path.exists(path) else []
    return []

def get_active():
    try:
        if working_directory:
            path = os.path.join(working_directory, '.genolearn')
            if os.path.exists(path):
                import json
                with open(path) as f:
                    return json.load(f)
    except:
        ...